;; Generated package description from svg-clock.el  -*- no-byte-compile: t -*-
(define-package "svg-clock" "1.2.0.20201201.222542" "Analog clock using Scalable Vector Graphics" '((svg "1.0") (emacs "27.0")) :url "https://elpa.gnu.org/packages/svg-clock.html" :keywords '("demo" "svg" "clock") :authors '(("Ulf Jasper" . "ulf.jasper@web.de")) :maintainer '("Ulf Jasper" . "ulf.jasper@web.de"))
